package net.knarcraft.bookswithoutborders.utility;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.config.BooksWithoutBordersConfig;
import net.knarcraft.bookswithoutborders.state.BookDirectory;
import org.bukkit.Bukkit;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.BookMeta;

import java.io.File;
import java.util.UUID;

import static net.knarcraft.bookswithoutborders.config.BooksWithoutBordersConfig.getSlash;
import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.cleanString;
import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.fixName;

/**
 * Helper class for getting abstract book information
 */
public final class BookHelper {

    private BookHelper() {

    }

    /**
     * Converts the author of a book from UUID if necessary
     *
     * @param author <p>The author string</p>
     * @return <p>The author string, converted if it was a UUID</p>
     */
    public static String authorFromUUID(String author) {
        try {
            UUID authorID = UUID.fromString(author);
            Player player = Bukkit.getPlayer(authorID);
            if (player != null) {
                author = player.getName();
            }
        } catch (IllegalArgumentException ignored) {
        }
        return author;
    }

    /**
     * Gets the file path of the selected book directory
     *
     * @param bookDirectory <p>The book directory to get (ENCRYPTED is not supported here)</p>
     * @param sender        <p>The command sender trying to get the directory</p>
     * @return <p>The path of the directory, or null if not possible to get</p>
     */
    public static File getBookDirectoryPath(BookDirectory bookDirectory, CommandSender sender) {
        String bookFolderString = getBookDirectoryPathString(bookDirectory, sender);
        if (bookFolderString == null) {
            return null;
        }
        return new File(bookFolderString);
    }

    /**
     * Gets the string path of the selected book directory
     *
     * @param bookDirectory <p>The book directory to get (ENCRYPTED is not supported here)</p>
     * @param sender        <p>The command sender trying to get the directory</p>
     * @return <p>The path of the directory, or null if not possible to get</p>
     */
    public static String getBookDirectoryPathString(BookDirectory bookDirectory, CommandSender sender) {
        String folder = null;
        String bookFolder = BooksWithoutBordersConfig.getBookFolder();
        if (bookDirectory == BookDirectory.PUBLIC) {
            folder = bookFolder;
        } else if (bookDirectory == BookDirectory.PLAYER && sender instanceof Player player) {
            folder = bookFolder + player.getUniqueId() + getSlash();
        }
        return folder;
    }

    /**
     * Increases the generation of the given book, if necessary
     *
     * @param bookItem <p>The book item to increase the generation of</p>
     */
    public static void increaseGeneration(ItemStack bookItem) {
        BookMeta bookMeta = (BookMeta) bookItem.getItemMeta();
        if (BooksWithoutBordersConfig.changeGenerationOnCopy() && bookMeta != null) {
            bookMeta.setGeneration(BookHelper.getNextGeneration(bookMeta.getGeneration()));
            bookItem.setItemMeta(bookMeta);
        }
    }

    /**
     * Gets the next generation of the given book
     *
     * <p>If an original book is given, this will yield a copy of the original. If a copy of original is given, this
     * will yield a copy of a copy. In all other cases, the generation will stay the same</p>
     *
     * @param currentGeneration <p>The current generation of the book</p>
     * @return <p>The next generation of the book</p>
     */
    public static BookMeta.Generation getNextGeneration(BookMeta.Generation currentGeneration) {
        if (currentGeneration == null) {
            return BookMeta.Generation.COPY_OF_ORIGINAL;
        }
        return switch (currentGeneration) {
            case ORIGINAL -> BookMeta.Generation.COPY_OF_ORIGINAL;
            case COPY_OF_ORIGINAL -> BookMeta.Generation.COPY_OF_COPY;
            default -> currentGeneration;
        };
    }

    /**
     * Gets the file name of the given book
     *
     * @param book   <p>The book to get the file of</p>
     * @param player <p>The player trying to do something with the book</p>
     * @return <p>The book file</p>
     */
    public static String getBookFile(BookMeta book, Player player, boolean isPublic) {
        String titleAuthorSeparator = BooksWithoutBordersConfig.getTitleAuthorSeparator();
        String bookName;
        if (book.hasTitle()) {
            bookName = book.getTitle();
        } else {
            bookName = "Untitled";
        }

        String authorName;
        if ((!book.hasAuthor() || isAuthor(player.getName(), book.getAuthor())) && !isPublic) {
            //Store as unique id to account for name changes
            authorName = player.getUniqueId().toString();
        } else if (!book.hasAuthor()) {
            authorName = player.getName();
        } else {
            authorName = book.getAuthor();
        }

        return fixName(cleanString(bookName + titleAuthorSeparator + authorName), false);
    }

    /**
     * Checks whether the given player is the author of a given book
     *
     * @param player <p>The player to check</p>
     * @param book   <p>The book to check</p>
     * @return <p>True if the player is not the book's author</p>
     */
    public static boolean isNotAuthor(Player player, BookMeta book) {
        if (isAuthor(player.getName(), book.getAuthor())) {
            return false;
        } else {
            BooksWithoutBorders.sendErrorMessage(player,
                    "You must be the author of this book to use this command!");
            return true;
        }
    }

    /**
     * Gets whether the given player name is equal to the given book author
     *
     * @param playerName <p>The player name to check</p>
     * @param author     <p>The author to check</p>
     * @return <p>True if the player is the author</p>
     */
    private static boolean isAuthor(String playerName, String author) {
        playerName = InputCleaningHelper.cleanString(playerName);
        return author != null && playerName.equalsIgnoreCase(InputCleaningHelper.cleanString(author));
    }

}
